__all__ = ['saga_step_handler', 'no_response_saga_step_handler',
           'send_saga_response']


import functools
import logging
from dataclasses import dataclass, asdict

import typing

import celery
from celery import Celery, Task

from .utils import success_task_name, failure_task_name, serialize_saga_error

logger = logging.getLogger(__name__)


def send_saga_response(celery_app: Celery,
                       response_task_name: str,
                       response_queue_name: str,
                       saga_id: int,
                       payload):  # assuming payload is a @dataclass
    return celery_app.send_task(
        response_task_name,
        args=[
            saga_id,
            payload
        ],
        queue=response_queue_name
    )


def _saga_step_handler(response_queue: typing.Union[str, None]):
    """
    Apply this decorator between @task and actual task handler.

    @command_handlers_celery_app.task(bind=True, name=verify_consumer_details_message.TASK_NAME)
    @saga_handler(CREATE_ORDER_SAGA_RESPONSE_QUEUE)

    Note: it's important to set bind=True in @task
      because @saga_handler will need access to celery task instance
    """
    def inner(func):
        @functools.wraps(func)
        def wrapper(celery_task: Task, saga_id: int, payload: dict):
            try:
                response_payload = func(celery_task, saga_id, payload)  # type: typing.Union[dict, None]
                # use convention response task name
                task_name = success_task_name(celery_task.name)
            except BaseException as exc:
                # let Celery handle retries
                if isinstance(exc, celery.exceptions.Retry):
                    raise

                logger.exception(exc)

                # serialize error in a unified way
                response_payload = asdict(serialize_saga_error(exc))
                # use convention response task name
                task_name = failure_task_name(celery_task.name)

            if response_queue:
                send_saga_response(celery_task.app,
                                   task_name,
                                   response_queue,
                                   saga_id,
                                   response_payload)
        return wrapper

    return inner


def saga_step_handler(response_queue: str):
    """
    Compensatable saga step assumed.
    For retriable steps, use corresponding decorator

    It's also assumed that you will use this decorator with
     @task decorator, see docstring for _saga_step_handler
    """
    return _saga_step_handler(response_queue)


no_response_saga_step_handler = _saga_step_handler(response_queue=None)
"""
It's assumed that you will use this decorator with
 @task decorator, see docstring for _saga_step_handler
"""





__all__ = ['AbstractSagaStateRepository', 'StatefulSaga']

import abc

from celery import Celery, Task

from .utils import success_task_name, failure_task_name
from .base_saga import BaseSaga, BaseStep
from .async_saga import AsyncSaga, AsyncStep


class AbstractSagaStateRepository(abc.ABC):
    @abc.abstractmethod
    def get_saga_state_by_id(self, saga_id: int) -> object:
        raise NotImplementedError

    @abc.abstractmethod
    def update_status(self, saga_id: int, status: str) -> object:
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, saga_id: int, **fields_to_update: str) -> object:
        raise NotImplementedError

    @abc.abstractmethod
    def on_step_failure(self, saga_id: int, failed_step: BaseStep, initial_failure_payload: dict) -> object:
        pass


class StatefulSaga(AsyncSaga, abc.ABC):
    """
    Note this class assumes sqlalchemy-mixins library is used.
    Use it rather as an example
    """
    saga_state_repository: AbstractSagaStateRepository = None
    _saga_state = None  # cached SQLAlchemy instance

    def __init__(self, saga_state_repository: AbstractSagaStateRepository, celery_app: Celery, saga_id: int):
        self.saga_state_repository = saga_state_repository
        super().__init__(celery_app, saga_id)

    @property
    def saga_state(self):
        if not self._saga_state:
            self._saga_state = self.saga_state_repository.get_saga_state_by_id(self.saga_id)

        return self._saga_state

    def run_step(self, step: BaseStep):
        self.saga_state_repository.update_status(self.saga_id, status=f'{step.name}.running')
        super().run_step(step)

    def compensate_step(self, step: BaseStep, initial_failure_payload: dict):
        self.saga_state_repository.update_status(self.saga_id, status=f'{step.name}.compensating')
        super().compensate_step(step, initial_failure_payload)
        self.saga_state_repository.update_status(self.saga_id, status=f'{step.name}.compensated')

    def on_step_success(self, step: AsyncStep, *args, **kwargs):
        self.saga_state_repository.update_status(self.saga_id, status=f'{step.name}.succeeded')
        super().on_async_step_success(step, *args, **kwargs)

    def on_step_failure(self, failed_step: AsyncStep, payload: dict):
        self.saga_state_repository.update_status(self.saga_id, status=f'{failed_step.name}.failed')
        super().on_async_step_failure(failed_step, payload)

    def on_saga_success(self):
        super().on_saga_success()
        self.saga_state_repository.update_status(self.saga_id, 'succeeded')

    def on_saga_failure(self, *args, **kwargs):
        super().on_saga_failure(*args, **kwargs)
        self.saga_state_repository.update_status(self.saga_id, 'failed')

    def compensate(self, failed_step: BaseStep,
                   initial_failure_payload: dict = None):
        self.saga_state_repository.on_step_failure(self.saga_id, failed_step, initial_failure_payload)
        super().compensate(failed_step, initial_failure_payload)

    @classmethod
    def register_async_step_handlers(cls,
                                     saga_state_repository: AbstractSagaStateRepository,
                                     celery_app: Celery):
        # noinspection PyTypeChecker
        dummy_saga_instance = cls(None, None, None)

        for step in dummy_saga_instance.async_steps:
            cls.register_success_handler_for_step(saga_state_repository,
                                                  celery_app, step)
            cls.register_failure_handler_for_step(saga_state_repository,
                                                  celery_app, step)

    @classmethod
    def register_success_handler_for_step(cls,
                                          saga_state_repository: AbstractSagaStateRepository,
                                          celery_app: Celery, step: AsyncStep):
        def on_success_handler(celery_task: Task, saga_id: int, payload: dict):
            saga = cls(saga_state_repository=saga_state_repository,
                       celery_app=celery_app, saga_id=saga_id)

            step_ = saga.get_async_step_by_success_task_name(celery_task.name)
            saga.on_async_step_success(step_, payload)

        celery_app.task(
            name=success_task_name(step.base_task_name),
            bind=True
        )(on_success_handler)

    @classmethod
    def register_failure_handler_for_step(cls, saga_state_repository: AbstractSagaStateRepository, celery_app: Celery, step: AsyncStep):

        def on_failure_handler(celery_task: Task, saga_id: int, payload: dict):
            saga = cls(saga_state_repository, celery_app, saga_id)

            step_ = saga.get_async_step_by_failure_task_name(celery_task.name)
            saga.on_async_step_failure(step_, payload)

        celery_app.task(
            name=failure_task_name(step.base_task_name),
            bind=True
        )(on_failure_handler)



__all__ = ['auto_retry_then_reraise', 'close_sqlalchemy_db_connection_after_celery_task_ends']

import functools

from celery import Task
from celery.exceptions import MaxRetriesExceededError
from celery.signals import task_postrun


def auto_retry_then_reraise(max_retries: int = 3, **retry_kwargs):
    """
    Retry max_retries times.
    If all retries failed, reraise initially risen error.

    Apply this decorator after Celery's @task decorator.
    For example:
        @celery_app.task(bind=True)
        @reraise_exception_after_max_retries_exceeded(max_retries=5)

    Note: it's important to set bind=True in @task
      because this decorator will need access to celery task instance
    """
    def inner(func):

        @functools.wraps(func)
        def wrapper(self: Task, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except Exception as exc:
                try:
                    raise self.retry(exc=exc, max_retries=max_retries, **retry_kwargs)
                except MaxRetriesExceededError:
                    raise exc

        return wrapper
    return inner


def close_sqlalchemy_db_connection_after_celery_task_ends(sqlalchemy_session):
    @task_postrun.connect
    def close_session(*args, **kwargs):
        sqlalchemy_session.remove()



"""
This module contains AsyncSaga and AsyncSteps which are the very heart
 of saga pattern in microservices.

AsyncSaga includes class methods that allow registering Celery tasks
 that correspond to responses from Saga Handler services

For example, when we send 'create_restaurant_ticket' task to Restaurant service,
 it will respond to message broker with Celery tasks like
   'create_restaurant_ticket.response.success' or
   'create_restaurant_ticket.reesponse.failure'

Saga Orchestrator has its own Celery worker which listen to 'response'
  Celery tasks and launch next step or rollback saga on failure.
See AsyncSaga.register_async_step_handlers for more details.

"""

__all__ = ['AsyncSaga', 'AsyncStep']

import logging
import typing

from celery import Celery, Task

from .base_saga import BaseSaga, BaseStep
from .utils import success_task_name, failure_task_name, NO_ACTION


logger = logging.getLogger(__name__)


class AsyncStep(BaseStep):
    def __init__(self,
                 base_task_name: str,
                 queue: str,
                 on_success: typing.Callable = NO_ACTION,
                 on_failure: typing.Callable = NO_ACTION,
                 *args, **kwargs
                 ):
        self.base_task_name = base_task_name
        self.queue = queue
        self.on_success = on_success
        self.on_failure = on_failure

        super().__init__(*args, **kwargs)


class AsyncSaga(BaseSaga):
    """
    Saga that has integration with Celery
    """
    celery_app: Celery = None

    def __init__(self, celery_app: Celery, *args, **kwargs):
        self.celery_app = celery_app
        super().__init__(*args, **kwargs)

    def on_async_step_success(self, step: AsyncStep, payload: dict):
        logger.info(f'Saga {self.saga_id}: '
                    f'running on_success for "{step.name}" step')

        step.on_success(step, payload)

        if self.step_is_last(step):
            self.on_saga_success()
        else:
            next_step = self._get_next_step(step)
            self.execute(next_step)

    def on_async_step_failure(self, step: AsyncStep, payload: dict):
        logger.info(f'Saga {self.saga_id}: '
                    f'running on_failure for "{step.name}" step')

        step.on_failure(step, payload)
        self.compensate(step, payload)

    @property
    def async_steps(self) -> typing.List[AsyncStep]:
        return [step for step in self.steps if isinstance(step, AsyncStep)]

    def get_async_step_by_success_task_name(self, success_task_name_: str) -> AsyncStep:
        for step in self.async_steps:
            if success_task_name(step.base_task_name) == success_task_name_:
                return step

        raise KeyError(f'no step found with success task name {success_task_name_}')

    def get_async_step_by_failure_task_name(self, failure_task_name_: str) -> AsyncStep:
        for step in self.async_steps:
            if failure_task_name(step.base_task_name) == failure_task_name_:
                return step

        raise KeyError(f'no step found with failure task name {failure_task_name_}')

    @classmethod
    def register_async_step_handlers(cls, celery_app: Celery):
        # noinspection PyTypeChecker
        dummy_saga_instance = cls(None, None)

        for step in dummy_saga_instance.async_steps:
            cls.register_success_handler_for_step(celery_app, step)
            cls.register_failure_handler_for_step(celery_app, step)

    @classmethod
    def register_success_handler_for_step(cls, celery_app: Celery, step: AsyncStep):
        def on_success_handler(celery_task: Task, saga_id: int, payload: dict):
            saga = cls(celery_app=celery_app, saga_id=saga_id)

            step_ = saga.get_async_step_by_success_task_name(celery_task.name)
            saga.on_async_step_success(step_, payload)

        celery_app.task(
            name=success_task_name(step.base_task_name),
            bind=True
        )(on_success_handler)

    @classmethod
    def register_failure_handler_for_step(cls, celery_app: Celery, step: AsyncStep):

        def on_failure_handler(celery_task: Task, saga_id: int, payload: dict):
            saga = cls(celery_app, saga_id)

            step_ = saga.get_async_step_by_failure_task_name(celery_task.name)
            saga.on_async_step_failure(step_, payload)

        celery_app.task(
            name=failure_task_name(step.base_task_name),
            bind=True
        )(on_failure_handler)

    def send_message_to_other_service(self, step: AsyncStep, payload: dict, task_name: str = None):
        """
        Helper for sending Celery tasks to Async Handler Services
        """
        task_result = self.celery_app.send_task(
            task_name or step.base_task_name,
            args=[
                self.saga_id,
                payload
            ],
            queue=step.queue
        )

        return task_result.id



from setuptools import setup


def requirements():
    import os
    filename = os.path.join(os.path.dirname(__file__), 'requirements.txt')
    return [line.rstrip('\n') for line in open(filename).readlines()]


setup(name='saga_framework',
      version='0.1',
      description='Mini-framework that helps implementing Saga pattern in microservices',
      url='https://github.com/absent1706/saga-framework',
      download_url='https://github.com/absent1706/saga-framework/archive/master.tar.gz',
      author='Alexander Litvinenko',
      author_email='litvinenko1706@gmail.com',
      license='MIT',
      packages=['saga_framework'],
      install_requires=requirements(),
      keywords=['microservices', 'saga'],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
      ]
)



__all__ = ['success_task_name', 'failure_task_name', 'SagaErrorPayload',
           'format_exception_as_python_does', 'serialize_saga_error',
           'NO_ACTION']

import traceback
from dataclasses import dataclass


NO_ACTION = lambda *args: None


def success_task_name(task_name: str):
    return f'{task_name}.response.success'


def failure_task_name(task_name: str):
    return f'{task_name}.response.failure'


@dataclass
class SagaErrorPayload:
    type: str
    message: str
    module: str
    traceback: str


def format_exception_as_python_does(e: BaseException):
    # taken from https://stackoverflow.com/a/35498685
    return traceback.format_exception(type(e), e, e.__traceback__)


def serialize_saga_error(exc: BaseException) -> SagaErrorPayload:
    exctype = type(exc)
    return SagaErrorPayload(
        type=getattr(exctype, '__qualname__', exctype.__name__),
        message=str(exc),
        module=exctype.__module__,
        traceback=format_exception_as_python_does(exc)
    )




__all__ = ['BaseStep', 'SyncStep', 'BaseSaga', 'NO_ACTION']

import logging
import typing
from abc import ABC
from dataclasses import asdict

from .utils import serialize_saga_error, \
    format_exception_as_python_does, NO_ACTION

logger = logging.getLogger(__name__)


class BaseStep(ABC):
    def __init__(self,
                 name: str,
                 action: typing.Callable = NO_ACTION,
                 compensation: typing.Callable = NO_ACTION,
                 ):
        self.name = name
        self.action = action
        self.compensation = compensation


class SyncStep(BaseStep):
    pass


class BaseSaga:
    saga_id: int = None
    steps: typing.List[BaseStep] = None

    def __init__(self, saga_id: int):
        self.saga_id = saga_id

    def get_first_step(self) -> BaseStep:
        return self.steps[0]

    def get_step_by_name(self, step_name: str) -> BaseStep:
        for step in self.steps:
            if step.name == step_name:
                return step

        raise KeyError(f'no step found with name {step_name}')

    def _get_step_index(self, step: BaseStep) -> int:
        for i in range(len(self.steps)):
            if self.steps[i].name == step.name:
                return i

        raise IndexError(f'step wasn\'t found')

    def _get_next_step(self, step: typing.Union[BaseStep, None]) -> typing.Union[BaseStep, None]:
        if not step:
            return self.steps[0]

        step_index = self._get_step_index(step)

        its_last_step = (step_index == len(self.steps) - 1)

        if its_last_step:
            return None
        else:
            return self.steps[step_index + 1]

    def _get_previous_step(self, step: typing.Union[BaseStep, None]) -> typing.Union[BaseStep, None]:
        step_index = self._get_step_index(step)

        its_first_step = (step_index == 0)

        if its_first_step:
            return None
        else:
            return self.steps[step_index - 1]

    def step_is_last(self, step: BaseStep):
        return step == self.steps[-1]

    def run_step(self, step: BaseStep):
        logger.info(f'Saga {self.saga_id}: running "{step.name}" step')
        step.action(step)

    def compensate_step(self, step: BaseStep, initial_failure_payload: dict):
        logger.info(f'Saga {self.saga_id}: '
                    f'compensating "{step.name}" step')
        step.compensation(step)

    def compensate(self, failed_step: BaseStep, initial_failure_payload: dict = None):
        try:
            step = self._get_previous_step(failed_step)
            while step:
                self.compensate_step(step, initial_failure_payload)
                step = self._get_previous_step(step)

            self.on_saga_failure(failed_step, initial_failure_payload)

        except BaseException as exception:
            self.on_compensation_failure(
                initially_failed_step=failed_step,
                initial_failure_payload=initial_failure_payload,
                compensation_failed_step=failed_step,
                compensation_exception=exception
            )

    def execute(self, starting_step: BaseStep = None):
        if starting_step is None:
            starting_step = self.steps[0]

        step = starting_step
        need_to_run_next_step = True
        exception = None

        while step and need_to_run_next_step:
            # noinspection PyBroadException
            try:
                self.run_step(step)

            except BaseException as exc:
                exception = exc
                break

            # After running a step, we will run next one if current step was sync
            # For AsyncStep's, we firstly wait for on_success event from step handlers
            #  and only then continue saga (see on_async_step_success method)
            need_to_run_next_step = isinstance(step, SyncStep)
            if need_to_run_next_step:
                step = self._get_next_step(step)

        # if error occured, compensate saga
        if exception:
            self.compensate(
                step,
                initial_failure_payload=asdict(serialize_saga_error(exception))
            )
        # if we ended on a last step, run on_saga_success
        elif step is None:
            self.on_saga_success()

    def on_saga_success(self):
        """
        This method runs when saga is fully completed with success
        """

        logger.info(f'Saga {self.saga_id} succeeded')

    def on_saga_failure(self, failed_step: BaseStep, initial_failure_payload: dict):
        """
        This method runs when saga is failed (after all compensations finished)
        """
        logger.info(f'Saga {self.saga_id} failed on "{failed_step.name}" step. \n'
                    f'Failure details: {initial_failure_payload}')

    def on_compensation_failure(self, initially_failed_step: BaseStep,
                                initial_failure_payload: dict,
                                compensation_failed_step: BaseStep,
                                compensation_exception: BaseException):
        """
        This method runs when compensation step unexpectedly failed,
          i.e. saga wasn't able to successfully rollback
        """
        logger.info(f'Saga {self.saga_id} failed while compensating "{compensation_failed_step.name}" step.\n'
                    f'Error details: {format_exception_as_python_does(compensation_exception)} \n \n'
                    f'Initial failure details: {initial_failure_payload}')




import typing

import asyncapi

from . import success_task_name, failure_task_name, SagaErrorPayload

fake_asyncapi_servers = {'development': asyncapi.Server(
    url='localhost',
    protocol=asyncapi.ProtocolType.REDIS,
    description="Don't look at this server description, it's just some fake needed to generate AsyncAPI docs"
)}


def message_to_channel(message: asyncapi.Message, response: asyncapi.Message = None, publish_made_first=False, description: str = None) -> typing.Tuple[str, asyncapi.Channel]:
    if publish_made_first:
        first_action, second_action = 'publish', 'subscribe'
    else:
        first_action, second_action = 'subscribe', 'publish'

    channel_kwargs = {
        'description': description,
        first_action: asyncapi.Operation(
            message=message,
        )
    }
    if response:
        channel_kwargs[second_action] = asyncapi.Operation(
            message=response,
        )

    return message.name, asyncapi.Channel(**channel_kwargs)


def message_to_component(message: asyncapi.Message) -> typing.Tuple[str, asyncapi.Message]:
    return message.name, message


def asyncapi_components_from_asyncapi_channels(channels: typing.Iterable[asyncapi.Channel]):
    """
    Allows to more easy generate AsyncApi docs (components sections)

    :param channels:
    :return:
    """

    messages = list()
    for channel in channels:
        if channel.publish and channel.publish.message:
            messages.append(channel.publish.message)
        if channel.subscribe and channel.subscribe.message:
            messages.append(channel.subscribe.message)

    components = [message_to_component(message) for message in messages]
    return asyncapi.Components(messages=dict(components))

#############


def asyncapi_message_for_success_response(base_task_name: str,
                                          title: str = None,
                                          summary: str = None,
                                          payload_dataclass: object = None):
    return asyncapi.Message(
        name=success_task_name(base_task_name),
        title=title,
        summary=summary,
        payload=payload_dataclass
    )


def asyncapi_message_for_failure_response(base_task_name: str,
                                          title: str = None,
                                          summary: str = None,
                                          payload_dataclass: type = SagaErrorPayload):
    return asyncapi.Message(
        name=failure_task_name(base_task_name),
        title=title,
        summary=summary,
        payload=payload_dataclass
    )
